#define _XOPEN_SOURCE 700

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/MainW.h>
#include <Xm/IconG.h>
#include <Xm/Container.h>
#include <Xm/ScrolledW.h>

#include <Xmd/Icon.h>
#include <Xmd/Exec.h>
#include <Xmd/Buffer.h>
#include <Xmd/StringMap.h>

#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include <dirent.h>
#include <libgen.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "icons/icon.xbm"
#include "icons/directory.xbm"
#include "icons/directory-small.xbm"
#include "icons/file.xbm"
#include "icons/file-small.xbm"
#include "icons/link.xbm"
#include "icons/link-small.xbm"
#include "icons/block.xbm"
#include "icons/block-small.xbm"
#include "icons/character.xbm"
#include "icons/character-small.xbm"
#include "icons/socket.xbm"
#include "icons/socket-small.xbm"
#include "icons/pipe.xbm"
#include "icons/pipe-small.xbm"

static XtAppContext application;
static XmdStringMap *trailFiles;
static XmdStringMap *viewFiles;
static Widget       topLevel;
static Widget       viewContainer;
static Widget       trailContainer;
static Pixmap       iconPixmap;
static Pixmap       iconsLarge[7] = { 0 };
static Pixmap       iconsSmall[7] = { 0 };

typedef enum {
	FileKindFile,
	FileKindDirectory,
	FileKindLink,
	FileKindBlock,
	FileKindCharacter,
	FileKindSocket,
	FileKindPipe
} FileKind;

typedef struct {
	Widget   widget;
	FileKind kind;
	String   name;
	String   label;
	String   path;
} File;

void     loadIconPixmaps  (void);
int      refreshDirectory ();
int      makeTrailIn      (ConstString path);
int      makeFilesIn      (ConstString directory, Bool showHidden);
int      changeDirectory  (ConstString path);
void     handleActivate   (Widget, XtPointer, XtPointer);
FileKind fileKind         (ConstString path);
File *   FileNew          (XmdStringMap *set, Widget container, String name, ConstString path);
File *   FileOfWidget     (Widget widget);
void     FileDelete       (XmdStringMap *set, ConstString name);
void     FileFree         (File *this);
void     FileSetLabel     (File *this, ConstString label);

int main (int argc, char *argv[]) {
	trailFiles = XmdStringMapNew();
	viewFiles = XmdStringMapNew();

	topLevel = XtVaAppInitialize (
		&application, "FileBrowser",
		NULL, 0,
		&argc, argv,
		NULL,
		XmNtitle, "File Browser",
		XmNiconName, "File Browser",
		XmNwidth,    400,
		XmNheight,   300,
		NULL);
	Widget window = XtVaCreateManagedWidget (
		"window", xmMainWindowWidgetClass, topLevel,
		NULL);

	/* menu bar */
	XmString fileString   = XmStringCreateLocalized("File");
	XmString helpString   = XmStringCreateLocalized("Help");
	Widget menuBar = XmVaCreateSimpleMenuBar (
		window, "menuBar",
    		XmVaCASCADEBUTTON, fileString,   'F',
		XmVaCASCADEBUTTON, helpString,   'H',
		NULL);
	XmStringFree(fileString);
	XmStringFree(helpString);
	XtManageChild(menuBar);

	/* layout form */
	Widget layout = XtVaCreateWidget (
		"layout", xmFormWidgetClass, window,
		XmNorientation, XmVERTICAL,
		NULL);
	iconPixmap = XmdLoadBitmapIcon(layout, icon);
	XtVaSetValues (
		topLevel,
		XmNiconPixmap, iconPixmap,
		NULL);

	/* trail container */
	Widget trailScroller = XtVaCreateWidget (
		"trailScroller", xmScrolledWindowWidgetClass, layout,
		XmNscrollingPolicy,         XmAUTOMATIC,
		XmNscrolledWindowChildType, XmSCROLL_HOR,
		XmNleftAttachment,   XmATTACH_FORM,
		XmNtopAttachment,    XmATTACH_FORM,
		XmNrightAttachment,  XmATTACH_FORM,
		NULL);
	trailContainer = XtVaCreateManagedWidget (
		"trailContainer", xmContainerWidgetClass, trailScroller,
		XmNlayoutType,          XmOUTLINE,
		XmNentryViewType,       XmSMALL_ICON,
		NULL);
	XtAddCallback (
		trailContainer, XmNdefaultActionCallback,
		handleActivate,
		NULL);

	/* file object container */	
	Widget viewScroller = XtVaCreateWidget (
		"viewScroller", xmScrolledWindowWidgetClass, layout,
		XmNscrollingPolicy,         XmAUTOMATIC,
		XmNscrolledWindowChildType, XmSCROLL_VERT,
		XmNleftAttachment,   XmATTACH_FORM,
		XmNtopAttachment,    XmATTACH_WIDGET,
		XmNtopWidget,        trailScroller,
		XmNrightAttachment,  XmATTACH_FORM,
		XmNbottomAttachment, XmATTACH_FORM,
		NULL);
	viewContainer = XtVaCreateManagedWidget (
		"viewContainer", xmContainerWidgetClass, viewScroller,
		XmNlayoutType,          XmSPATIAL,
		XmNspatialStyle,        XmGRID,
		XmNspatialIncludeModel, XmFIRST_FIT,
		XmNspatialResizeModel,  XmGROW_MINOR,
		NULL);
	XtAddCallback (
		viewContainer, XmNdefaultActionCallback,
		handleActivate,
		NULL);
	
	loadIconPixmaps();
	refreshDirectory();

	XtManageChild(viewScroller);
	XtManageChild(trailScroller);
	XtManageChild(layout);
	
	XtVaSetValues (
		window,
		XmNmenuBar,    menuBar,
		XmNworkWindow, layout,
		NULL);
	
	XtRealizeWidget(topLevel);
	XtAppMainLoop(application);
}

void loadIconPixmaps (void) {
	#define kind(kind, name) \
		iconsLarge[FileKind##kind] = XmdLoadBitmapIcon(viewContainer, name); \
		iconsSmall[FileKind##kind] = XmdLoadBitmapIcon(viewContainer, name##_small)
	kind(File,      file);
	kind(Directory, directory);
	kind(Link,      link);
	kind(Block,     block);
	kind(Character, character);
	kind(Socket,    socket);
	kind(Pipe,      pipe);
}

Bool freeFile (XmdStringMap *map, ConstString key, void *value, void *data) {
	(void)(map);
	(void)(key);
	(void)(data);
	File *this = value;
	FileFree(this);
	return True;
}

int refreshDirectory () {
	char directory[PATH_MAX];
	getcwd(directory, XtNumber(directory));
	XtVaSetValues (
		topLevel,
		XmNtitle, directory,
		XmNiconName, directory,
		NULL);
	
	int err = 0;
	XtUnmanageChild(trailContainer); {
		XmdStringMapIterate(trailFiles, freeFile, NULL);
		XmdStringMapFree(trailFiles);
		trailFiles = XmdStringMapNew();
		err |= makeTrailIn(directory);
	} XtManageChild(trailContainer);
	
	XtUnmanageChild(viewContainer); {
		XmdStringMapIterate(viewFiles, freeFile, NULL);
		XmdStringMapFree(viewFiles);
		viewFiles = XmdStringMapNew();
		err |= makeFilesIn(directory, True);
	} XtManageChild(viewContainer);
	return err;
}

int changeDirectory (ConstString path) {
	int err = chdir(path);
	if (err) return err;
	refreshDirectory();
	return 0;
}

void handleActivate (Widget container, XtPointer clientData, XtPointer callData) {
	(void)(container);
	(void)(clientData);
	XmContainerSelectCallbackStruct *event = callData;
	if (event->selected_item_count < 1) return;
	
	/* first, if a directory is selected, try to go into it */
	for (int index = 0; index < event->selected_item_count; index ++) {
		File *file = FileOfWidget(event->selected_items[index]);
		if (file == NULL) continue;
		if (file->kind == FileKindDirectory) {
			changeDirectory(file->path);
			return;
		}
	}
	
	/* if not, we worry about opening files and stuff */
	/* TODO */
}

FileKind fileKind (ConstString path) {
	struct stat info = { 0 };
	stat(path, &info);
	if (S_ISDIR (info.st_mode)) return FileKindDirectory;
	if (S_ISCHR (info.st_mode)) return FileKindCharacter;
	if (S_ISBLK (info.st_mode)) return FileKindBlock;
	if (S_ISFIFO(info.st_mode)) return FileKindPipe;
	if (S_ISLNK (info.st_mode)) return FileKindLink;
	if (S_ISSOCK(info.st_mode)) return FileKindSocket;
	return FileKindFile;
}

int makeTrailIn (ConstString cwd) {
	char name[PATH_MAX] = { 0 }; Cardinal nameIndex = 0;
	char path[PATH_MAX] = { 0 }; Cardinal pathIndex = 0;
	
	while (True) {
		char ch = *(cwd ++);

		if (nameIndex >= XtNumber(name)) return -1;
		if (pathIndex >= XtNumber(path)) return -1;
		
		name[   nameIndex] = ch;
		name[++ nameIndex] = 0;
		path[   pathIndex] = ch;
		path[++ pathIndex] = 0;
		
		if (ch == '/' || ch == 0) {
			if (nameIndex == 0) {
				name[0] = '/';
				name[1] = 0;
			}
			
			FileNew(trailFiles, trailContainer, name, path);
			name[(nameIndex = 0)] = 0;
			if (ch == 0) return 0;
		}
	}
}

int makeFilesIn (ConstString directory, Bool showHidden) {
	struct dirent **entries = NULL;
	int entriesCount = scandir(directory, &entries, NULL, alphasort);
	if (entriesCount < 0) {
		return -1;
	}
	for (int index = 0; index < entriesCount; index ++) {
		struct dirent *entry = entries[index];
		if (entry->d_name[0] == '.' && !showHidden) {
			XtFree((char *)(entry));
			continue;
		}
		char filePath[PATH_MAX];
		snprintf (
			filePath, XtNumber(filePath),
			"%s/%s", directory, entry->d_name);
		
		String fileName = XtNewString(entry->d_name);
		XtFree((char *)(entry));
		
		FileNew(viewFiles, viewContainer, fileName, filePath);
		XtFree(fileName);
	}
	XtFree((char *)(entries));
	return 0;
}

File *FileNew (XmdStringMap *set, Widget container, String name, ConstString path) {
	File *this = XtNew(File);
	this->label      = NULL;
	this->name       = XtNewString(name);
	this->path       = XtNewString(path);
	XmString string = XmStringCreateLocalized(name);
	this->widget = XtVaCreateManagedWidget (
		"file", xmIconGadgetClass, container,
		XmNalignment,   XmALIGNMENT_BEGINNING,
		XmNlabelString, string,
		XmNuserData,    (XtPointer)(this),
		NULL);
	XmStringFree(string);
	
	FileKind kind = fileKind(path);
	XtVaSetValues (
		this->widget,
		XmNlargeIconPixmap, iconsLarge[kind],
		XmNsmallIconPixmap, iconsSmall[kind],
		NULL);
	this->kind = kind;
	
	XmdStringMapSet(set, name, this);
	return this;
}

File *FileOfWidget (Widget widget) {
	File *file = NULL;
	XtVaGetValues(widget, XmNuserData, &file, NULL);
	return file;
}

void FileDelete (XmdStringMap *set, ConstString name) {
	FileFree(XmdStringMapSet(set, name, NULL));
}

void FileFree (File *this) {
	if (this == NULL) return;
	XtUnmanageChild(this->widget);
	XtDestroyWidget(this->widget);
	XtFree((char *)(this->name));
	XtFree((char *)(this->label));
	XtFree((char *)(this->path));
	XtFree((char *)(this));
}

void FileSetLabel (File *this, ConstString label) {
	XtFree(this->label);
	this->label = XtNewString(label);
	
	XmString string = NULL;
	if (label == NULL) {
		string = XmStringCreateLocalized(this->name);
	} else {
		string = XmStringCreateLocalized(this->label);
	}
	XtVaSetValues (
		this->widget,
		XmNlabelString, string,
		NULL);
	XmStringFree(string);
}
